@@ -33,29 +33,57 @@ __author_email__ = "biziqe@mathieu.fenniak.net"
 
 
 import sys
-# "Str" maintains compatibility with Python 2.x.
-# The next line is obfuscated like this so 2to3 won't change it.
+
+# See https://github.com/py-pdf/PyPDF2/issues/779
+from PyPDF2.errors import (  # noqa
+    STREAM_TRUNCATED_PREMATURELY,
+    PageSizeNotDefinedError,
+    PdfReadError,
+    PdfReadWarning,
+    PdfStreamError,
+    PyPdfError,
+)
+
 try:
-    import __builtin__ as builtins
-except ImportError:  # Py3
     import builtins
+    from typing import Dict
+except ImportError:  # Py2.7
+    import __builtin__ as builtins  # type: ignore
 
 
-if sys.version_info[0] < 3:
-    string_type = unicode
-    bytes_type = str
-else:
-    string_type = str
-    bytes_type = bytes
+xrange_fn = getattr(builtins, "xrange", range)
+_basestring = getattr(builtins, "basestring", str)
 
-Str = getattr(builtins, "basestring", str)
+bytes_type = type(bytes())  # Works the same in Python 2.X and 3.X
+string_type = getattr(builtins, "unicode", str)
+int_types = (int, long) if sys.version_info[0] < 3 else (int,)  # type: ignore  # noqa
+
+
+# Make basic type tests more consistent
+def isString(s):
+    """Test if arg is a string. Compatible with Python 2 and 3."""
+    return isinstance(s, _basestring)
+
+
+def isInt(n):
+    """Test if arg is an int. Compatible with Python 2 and 3."""
+    return isinstance(n, int_types)
+
+
+def isBytes(b):
+    """Test if arg is a bytes instance. Compatible with Python 2 and 3."""
+    import warnings
+
+    warnings.warn("PyPDF2.utils.isBytes will be deprecated", DeprecationWarning)
+    return isinstance(b, bytes_type)
 
 
-#custom implementation of warnings.formatwarning
 def formatWarning(message, category, filename, lineno, line=None):
-    file = filename.replace("/", "\\").rsplit("\\", 1)[1] # find the file name
+    """custom implementation of warnings.formatwarning"""
+    file = filename.replace("/", "\\").rsplit("\\", 1)[-1]  # find the file name
     return "%s: %s [%s:%s]\n" % (category.__name__, message, file, lineno)
 
+
 def readUntilWhitespace(stream, maxchars=None):
     """
     Reads non-whitespace characters and returns them.
@@ -71,34 +99,62 @@ def readUntilWhitespace(stream, maxchars=None):
             break
     return txt
 
+
 def readNonWhitespace(stream):
     """
     Finds and reads the next non-whitespace character (ignores whitespace).
     """
-    tok = b_(' ')
+    tok = WHITESPACES[0]
     while tok in WHITESPACES:
         tok = stream.read(1)
     return tok
 
+
 def skipOverWhitespace(stream):
     """
     Similar to readNonWhitespace, but returns a Boolean if more than
     one whitespace character was read.
     """
-    tok = b_(' ')
-    cnt = 0;
+    tok = WHITESPACES[0]
+    cnt = 0
     while tok in WHITESPACES:
         tok = stream.read(1)
-        cnt+=1
-    return (cnt > 1)
+        cnt += 1
+    return cnt > 1
+
 
 def skipOverComment(stream):
     tok = stream.read(1)
     stream.seek(-1, 1)
-    if tok == b_('%'):
-        while tok not in (b_('\n'), b_('\r')):
+    if tok == b_("%"):
+        while tok not in (b_("\n"), b_("\r")):
             tok = stream.read(1)
 
+
+def readUntilRegex(stream, regex, ignore_eof=False):
+    """
+    Reads until the regular expression pattern matched (ignore the match)
+    :raises PdfStreamError: on premature end-of-file
+    :param bool ignore_eof: If true, ignore end-of-line and return immediately
+    """
+    name = b_("")
+    while True:
+        tok = stream.read(16)
+        if not tok:
+            # stream has truncated prematurely
+            if ignore_eof:
+                return name
+            else:
+                raise PdfStreamError(STREAM_TRUNCATED_PREMATURELY)
+        m = regex.search(tok)
+        if m is not None:
+            name += tok[: m.start()]
+            stream.seek(m.start() - len(tok), 1)
+            break
+        name += tok
+    return name
+
+
 class ConvertFunctionsToVirtualList(object):
     def __init__(self, lengthFunction, getFunction):
         self.lengthFunction = lengthFunction
@@ -108,7 +164,11 @@ class ConvertFunctionsToVirtualList(object):
         return self.lengthFunction()
 
     def __getitem__(self, index):
-        if not isinstance(index, int):
+        if isinstance(index, slice):
+            indices = xrange_fn(*index.indices(len(self)))
+            cls = type(self)
+            return cls(indices.__len__, lambda idx: self[indices[idx]])
+        if not isInt(index):
             raise TypeError("sequence indices must be integers")
         len_self = len(self)
         if index < 0:
@@ -118,68 +178,73 @@ class ConvertFunctionsToVirtualList(object):
             raise IndexError("sequence index out of range")
         return self.getFunction(index)
 
+
 def RC4_encrypt(key, plaintext):
-    S = [i for i in range(256)]
+    S = list(range(256))
     j = 0
     for i in range(256):
         j = (j + S[i] + ord_(key[i % len(key)])) % 256
         S[i], S[j] = S[j], S[i]
     i, j = 0, 0
-    retval = b_("")
+    retval = []
     for x in range(len(plaintext)):
         i = (i + 1) % 256
         j = (j + S[i]) % 256
         S[i], S[j] = S[j], S[i]
         t = S[(S[i] + S[j]) % 256]
-        retval += b_(chr(ord_(plaintext[x]) ^ t))
-    return retval
+        retval.append(b_(chr(ord_(plaintext[x]) ^ t)))
+    return b_("").join(retval)
+
 
 def matrixMultiply(a, b):
-    return [[sum([float(i)*float(j)
-                  for i, j in zip(row, col)]
-                ) for col in zip(*b)]
-            for row in a]
+    return [
+        [sum([float(i) * float(j) for i, j in zip(row, col)]) for col in zip(*b)]
+        for row in a
+    ]
+
 
 def markLocation(stream):
     """Creates text file showing current location in context."""
     # Mainly for debugging
-    RADIUS = 5000
-    stream.seek(-RADIUS, 1)
-    outputDoc = open('PyPDF2_pdfLocation.txt', 'w')
-    outputDoc.write(stream.read(RADIUS))
-    outputDoc.write('HERE')
-    outputDoc.write(stream.read(RADIUS))
-    outputDoc.close()
-    stream.seek(-RADIUS, 1)
-
-class PyPdfError(Exception):
-    pass
-
-class PdfReadError(PyPdfError):
-    pass
-
-class PageSizeNotDefinedError(PyPdfError):
-    pass
-
-class PdfReadWarning(UserWarning):
-    pass
-
-class PdfStreamError(PdfReadError):
-    pass
+    radius = 5000
+    stream.seek(-radius, 1)
+    with open("PyPDF2_pdfLocation.txt", "wb") as output_fh:
+        output_fh.write(stream.read(radius))
+        output_fh.write(b"HERE")
+        output_fh.write(stream.read(radius))
+    stream.seek(-radius, 1)
 
 
-def b_(s):
-    if sys.version_info[0] < 3:
+if sys.version_info[0] < 3:
+
+    def b_(s):
         return s
-    else:
+
+else:
+    B_CACHE = {}  # type: Dict[str, bytes]
+
+    def b_(s):
+        bc = B_CACHE
+        if s in bc:
+            return bc[s]
         if type(s) == bytes:
             return s
         else:
-            return s.encode('latin-1')
+            try:
+                r = s.encode("latin-1")
+                if len(s) < 2:
+                    bc[s] = r
+                return r
+            except Exception:
+                r = s.encode("utf-8")
+                if len(s) < 2:
+                    bc[s] = r
+                return r
+
 
 def u_(s):
     if sys.version_info[0] < 3:
-        return unicode(s, 'unicode_escape')
+        return unicode(s, "unicode_escape")  # noqa
     else:
         return s
 
@@ -189,38 +254,58 @@ def str_(b):
         return b
     else:
         if type(b) == bytes:
-            return b.decode('latin-1')
+            return b.decode("latin-1")
         else:
             return b
 
+
 def ord_(b):
     if sys.version_info[0] < 3 or type(b) == str:
         return ord(b)
     else:
         return b
 
+
 def chr_(c):
     if sys.version_info[0] < 3:
         return c
     else:
         return chr(c)
 
+
 def barray(b):
     if sys.version_info[0] < 3:
         return b
     else:
         return bytearray(b)
 
+
 def hexencode(b):
     if sys.version_info[0] < 3:
-        return b.encode('hex')
+        return b.encode("hex")
     else:
         import codecs
-        coder = codecs.getencoder('hex_codec')
+
+        coder = codecs.getencoder("hex_codec")
         return coder(b)[0]
 
+
 def hexStr(num):
-    return hex(num).replace('L', '')
+    return hex(num).replace("L", "")
 
 
-WHITESPACES = [b_(x) for x in [' ', '\n', '\r', '\t', '\x00']]
+WHITESPACES = [b_(x) for x in [" ", "\n", "\r", "\t", "\x00"]]
+
+
+def paethPredictor(left, up, up_left):
+    p = left + up - up_left
+    dist_left = abs(p - left)
+    dist_up = abs(p - up)
+    dist_up_left = abs(p - up_left)
+
+    if dist_left <= dist_up and dist_left <= dist_up_left:
+        return left
+    elif dist_up <= dist_up_left:
+        return up
+    else:
+        return up_left
